local _G = _G
local MT = MacroToolkit
local AceConfig = MT.LS("AceConfig-3.0")
local AceConfigDialog = MT.LS("AceConfigDialog-3.0")
local AceDBOptions = MT.LS("AceDBOptions-3.0")
local L = MT.L
local CreateFrame, ipairs, pairs, string, tonumber = CreateFrame, ipairs, pairs, string, tonumber

--First visible frame
local function createMainPanel()
	local frame = CreateFrame("Frame", "MacroToolkitOptionsMain")
	local title = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
	local version = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
	local author = frame:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
	title:SetFormattedText("|T%s:%d|t %s", "Interface\\MacroFrame\\MacroFrame-Icon", 16, "Macro Toolkit")
	title:SetPoint("CENTER", frame, "CENTER", 0, 170)
	version:SetText(_G.GAME_VERSION_LABEL .. " " .. GetAddOnMetadata("MacroToolkit", "Version"))
	version:SetPoint("CENTER", frame, "CENTER", 0, 130)
	author:SetText(L["Author"] .. ": Deepac")
	author:SetPoint("CENTER", frame, "CENTER", 0, 100)
	return frame
end

local checkPanel = {
	order = 1,
	type = "group",
	name = _G.MAIN_MENU,
	args = {
		override = {
			order = 1,
			type = "toggle",
			width = "full",
			name = L["Override built in macro UI"],
			get = function() return MT.db.profile.override end,
			set = function(info, value)
					MT.db.profile.override = value
					if value then
						MT.showmacroframe = ShowMacroFrame
						ShowMacroFrame = function() MacroToolkitFrame:Show() end
					elseif MT.showmacroframe then ShowMacroFrame = MT.showmacroframe end
				end,
			},
		usecolour = {
			order = 2,
			type = "toggle",
			name = L["Use syntax highlighting"],
			width = "full",
			get = function() return MT.db.profile.usecolours end,
			set = function(info, value)
					MT.db.profile.usecolours = value
					if value then MacroToolkitFauxScrollFrame:Show()
					else MacroToolkitFauxScrollFrame:Hide() end
				end,
		},
		showerrors = {
			order = 3,
			type = "toggle",
			name = L["Display errors"],
			width = "full",
			get = function() return MT.db.profile.showerrors end,
			set = function(info, value)
					MT.db.profile.showerrors = value
					if value then
						MacroToolkitErrorBg:Show()
						MacroToolkitErrorScrollFrame:Show()
					else
						MacroToolkitErrorBg:Hide()
						MacroToolkitErrorScrollFrame:Hide()
					end
				end,
		},
		unknown = {
			order = 4,
			type = "toggle",
			name = L["Unknown parameter causes error"],
			width = "full",
			get = function() return MT.db.profile.unknown end,
			set = function(info, value)
					MT.db.profile.unknown = value
					if value then
						MacroToolkitErrorBg:Show()
						MacroToolkitErrorScrollFrame:Show()
					else
						MacroToolkitErrorBg:Hide()
						MacroToolkitErrorScrollFrame:Hide()
					end
				end,
		},
		replace = {
			order = 5,
			type = "toggle",
			name = L["Replace scripts with slash command"],
			desc = L["Replace known scripts with Macro Toolkit slash commands"],
			width = "full",
			get = function() return MT.db.profile.replacemt end,
			set  = function(info, value) MT.db.profile.replacemt  = value end,
		},
		spacer1 = {
			order = 6,
			type = "description",
			name = "\n",
		},
		scale = {
			order = 7,
			type = "range",
			name = _G.UI_SCALE,
			min = 0.25,
			max = 1.5,
			isPercent = true,
			step = 0.01,
			width = "double",
			get = function() return MT.db.profile.scale end,
			set = 
				function(info, value)
					MT.db.profile.scale = value
					if not (MacroToolkitFrame:IsShown() or MacroToolkitScriptFrame:IsShown() or MacroToolkitBuilderFrame:IsShown()) then
						MacroToolkitFrame:Show()
					end
					MacroToolkitFrame:SetPoint("BOTTOMLEFT", 50, 200)
					MacroToolkitScriptFrame:SetPoint("BOTTOMLEFT", 50, 200)
					MacroToolkitBuilderFrame:SetPoint("BOTTOMLEFT", 50, 200)
					MacroToolkitFrame:SetScale(value)
					MacroToolkitRestoreFrame:SetScale(value)
					MacroToolkitPopup:SetScale(value)
					MacroToolkitScriptFrame:SetScale(value)
					MacroToolkitBuilderFrame:SetScale(value)
					MT.db.profile.x = 50
					MT.db.profile.y = 200
				end,
		},
	},
}

local coloursPanel = {
	order = 2,
	type = "group",
	name = L["Syntax Highlighting"],
	args = {
		defaultcolour = {
			order = 1,
			type = "color",
			width = "full",
			name = _G.DEFAULT,
			get = function() return MT:HexToRGB(MT.db.profile.defaultcolour) end,
			set = function(info, r, g, b, a) MT.db.profile.defaultcolour = MT:RGBToHex(r, g, b, a) end
		},
		stringcolour = {
			order = 2,
			type = "color",
			width = "full",
			name = L["Text"],
			get = function() return MT:HexToRGB(MT.db.profile.stringcolour) end,
			set = function(info, r, g, b, a) MT.db.profile.stringcolour = MT:RGBToHex(r, g, b, a) end
		},
		emotecolour = {
			order = 3,
			type = "color",
			width = "full",
			name = L["Emotes"],
			get = function() return MT:HexToRGB(MT.db.profile.emotecolour) end,
			set = function(info, r, g, b, a) MT.db.profile.emotecolour = MT:RGBToHex(r, g, b, a) end
		},
		scriptcolour = {
			order = 4,
			type = "color",
			width = "full",
			name = L["Scripts"],
			get = function() return MT:HexToRGB(MT.db.profile.scriptcolour) end,
			set = function(info, r, g, b, a) MT.db.profile.scriptcolour = MT:RGBToHex(r, g, b, a) end
		},
		commandcolour = {
			order = 5,
			type = "color",
			width = "full",
			name = L["Commands"],
			get = function() return MT:HexToRGB(MT.db.profile.commandcolour) end,
			set = function(info, r, g, b, a) MT.db.profile.commandcolour = MT:RGBToHex(r, g, b, a) end
		},
		spellcolour = {
			order = 6,
			type = "color",
			width = "full",
			name = _G.SPELLS,
			get = function() return MT:HexToRGB(MT.db.profile.spellcolour) end,
			set = function(info, r, g, b, a) MT.db.profile.spellcolour = MT:RGBToHex(r, g, b, a) end
		},
		targetcolour = {
			order = 7,
			type = "color",
			width = "full",
			name = L["Targets"],
			get = function() return MT:HexToRGB(MT.db.profile.targetcolour) end,
			set = function(info, r, g, b, a) MT.db.profile.targetcolour = MT:RGBToHex(r, g, b, a) end
		},
		conditioncolour = {
			order = 8,
			type = "color",
			width = "full",
			name = L["Conditions"],
			get = function() return MT:HexToRGB(MT.db.profile.conditioncolour) end,
			set = function(info, r, g, b, a) MT.db.profile.conditioncolour = MT:RGBToHex(r, g, b, a) end
		},
		scriptcolour = {
			order = 9,
			type = "color",
			width = "full",
			name = _G.ERRORS,
			get = function() return MT:HexToRGB(MT.db.profile.errorcolour) end,
			set = function(info, r, g, b, a) MT.db.profile.errorcolour = MT:RGBToHex(r, g, b, a) end
		},
		itemcolour = {
			order = 10,
			type = "color",
			width = "full",
			name = _G.ITEMS,
			get = function() return MT:HexToRGB(MT.db.profile.itemcolour) end,
			set = function(info, r, g, b, a) MT.db.profile.itemcolour = MT:RGBToHex(r, g, b, a) end
		},
		mtcolour = {
			order = 11,
			type = "color",
			width = "full",
			name = L["Macro Toolkit slash commands"],
			get = function() return MT:HexToRGB(MT.db.profile.mtcolour) end,
			set = function(info, r, g, b, a) MT.db.profile.mtcolour = MT:RGBToHex(r, g, b, a) end
		},
	},
}

local function resetcolours()
	for k, v in pairs(MT.defaults.profile) do
		if string.find(k, "colour") then MT.db.profile[k] = v end
	end
	MT.ColoursFrame:Hide()
	MT.ColoursFrame:Show()
end

local function resetoptions()
	for k, v in pairs(MT.defaults.profile) do
		if not string.find(k, "colour") then MT.db.profile[k] = v end
	end
	if not MT.db.profile.usecolours then MacroToolkitFauxScrollFrame:Hide()
	else MacroToolkitFauxScrollFrame:Show() end
	if not MT.db.profile.showerrors then
		MacroToolkitErrorBg:Hide()
		MacroToolkitErrorScrollFrame:Hide()
	else
		MacroToolkitErrorBg:Show()
		MacroToolkitErrorScrollFrame:Show()
	end
	MT.OptionsFrame:Hide()
	MT.OptionsFrame:Show()
end

--Setup options
function MT:CreateOptions()
	local mainPanel = createMainPanel()
	mainPanel.name = "Macro Toolkit"
	InterfaceOptions_AddCategory(mainPanel)
	AceConfig:RegisterOptionsTable("MacroToolkitOptionsCheck", checkPanel)
	AceConfig:RegisterOptionsTable("MacroToolkitOptionsColours", coloursPanel)
	AceConfig:RegisterOptionsTable("MacroToolkitOptionsProfiles", AceDBOptions:GetOptionsTable(MT.db))
	MT.OptionsFrame = AceConfigDialog:AddToBlizOptions("MacroToolkitOptionsCheck", _G.MAIN_MENU, "Macro Toolkit")
	MT.OptionsFrame.default = function() resetoptions() end
	MT.ColoursFrame = AceConfigDialog:AddToBlizOptions("MacroToolkitOptionsColours", L["Syntax Highlighting"], "Macro Toolkit")
	MT.ColoursFrame.default = function() resetcolours() end
	AceConfigDialog:AddToBlizOptions("MacroToolkitOptionsProfiles", L["Profiles"], "Macro Toolkit")
end
